// BUMP, the imageboard porter
// ===========================
// -This software is designed to help out anon with imageboards

#include "Bump.h"

#include <algorithm>
#include <cstdio>
#include <cstdlib>
#include <fstream>
#include <iomanip>
#include <ios>
#include <iostream>
#include <iterator>
#include <map>
#include <sstream>
#include <string>
#include <tuple>
#include <utility>
#include <vector>

#include "muh_curl.h"

using std::begin;
using std::cerr;
using std::cout;
using std::end;
using std::endl;
using std::fstream;
using std::get;
using std::ifstream;
using std::ios;
using std::istringstream;
using std::make_tuple;
using std::map;
using std::ofstream;
using std::pair;
using std::printf;
using std::string;
using std::to_string;
using std::tuple;
using std::vector;

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

Page::Page()
{
}

Page::Page(const std::string& site_name,
           const std::string& board_name,
           const std::string& page_name,
           const fs::path& target_dir)
    : site_nm{site_name},
      board_nm{board_name},
      page_nm{page_name},
      targ_dir{target_dir}
{
  jsondb.parent = this;  // Error-reporting during load()

  // DESIGN: Future adaptation for sites that don't provide a JSON API
  // if (site_has_json(site_nm))
  load_json();
  // else {
  //   load_html();
  //   const auto comp_json{compile_html_to_json()};
  //   load_json(comp_json);
  // }
}

//------------------------------------------------------------------------------

// Get the live json and store it into the page's db
void Page::load_json()
{
  resp("json");
  jsondb.load(raw_json);
}

//------------------------------------------------------------------------------

// Store the board's live response inside the page
void Page::resp(const std::string& page_tp)
{
  const auto page_uri{muh_curl::muh_url(site_nm, board_nm, page_nm, page_tp)};

  if (page_tp == "json")
    raw_json = muh_curl::get_resp(page_uri).str();
  else
    raw_html = muh_curl::get_resp(page_uri).str();
}

//------------------------------------------------------------------------------

// Parse the raw json into the page's json db
void Page::Jsondb::load(const std::string& raw_json)
{
  Json::Reader reader{};

  if (!reader.parse(raw_json, db))
    exit_bad_status(raw_json);
}

//------------------------------------------------------------------------------

std::string lower_cp(const std::string& input)
{
  string res{input};
  std::transform(
      begin(res), end(res), begin(res), [](unsigned char c) -> unsigned char {
        return std::tolower(c);
      });
  return res;
}

void query_user(const std::string& site_nm,
                const std::string& board_nm,
                const std::string& page_nm,
                const std::string& err_msg = "")
{
  fprintf(stderr, "\nInput JSON data error: '%s'\n", err_msg.c_str());
  printf("Is   https://%s/%s/%s.json   supplying valid json?\n"
         "(this is the location BUMP is looking to for it's data)\n",
         site_nm.c_str(),
         board_nm.c_str(),
         page_nm.c_str());
}

// Search response for HTTP error codes to report, then exit.
void Page::Jsondb::exit_bad_status(const std::string& raw_resp) const
{
  // test:
  // cerr << "\n\nexit_bad_resp(): '" << raw_resp << "'\n\n";

  uint err_code{0};
  string err_msg{};
  vector<pair<string, uint>> err_codes{
      {"403 Forbidden", 403}, {"404 Not Found", 404}, {"502 Bad Gateway", 502}};

  for (const auto& [msg, code] : err_codes) {
    auto resp_it{lower_cp(raw_resp).find(lower_cp(msg))};
    if (resp_it != string::npos) {
      err_msg = msg;
      err_code = code;

      break;
    }
  }

  if (err_code) {
    query_user(parent->site_nm, parent->board_nm, parent->page_nm, err_msg);
    printf("HTTP status code: %u\n\n", err_code);
  }
  else {
    query_user(parent->site_nm, parent->board_nm, parent->page_nm);
    printf("\n\n");
  }

  std::exit(255);
}

//------------------------------------------------------------------------------

// Store all a thread's media files into it's local directory
void Page::grab_thrd_files() const
{
  const fs::path thumbs_dir{targ_dir / ".thumbs"};
  fs::create_directory(thumbs_dir);
  const fs::path media_dir{targ_dir / "media"};
  fs::create_directory(media_dir);

  grab_files(json()["files"], media_dir);

  const auto posts{json()["posts"]};
  for (const auto& post : posts)
    grab_files(post["files"], media_dir);
}

//------------------------------------------------------------------------------

// Save the thread's media files as needed
uint Page::grab_files(const Json::Value& files_arr,
                      const fs::path& media_dir) const
{
  uint file_cnt{0};

  for (const auto& file : files_arr) {
    const auto orig_file{file["originalName"].asString()};
    const fs::path media_file{trunc_filename(orig_file, media_dir)};

    // We never need to re-download a file once we have it locally.
    if (!fs::exists(media_file)) {
      const auto path_of_file{file["path"].asString()};
      auto file_url{muh_curl::muh_file_url(site_nm, path_of_file)};
      auto file_resp = muh_curl::get_resp(file_url);  // DESIGN: get_fresp()

      ofstream ofsb(media_file, std::ios::binary);
      ofsb << file_resp.rdbuf();
      ++file_cnt;
    }
  }

  return file_cnt;
}

//------------------------------------------------------------------------------

// Ensure the file name isn't too long
fs::path Page::trunc_filename(const std::string& orig_file,
                              const fs::path& media_dir) const
{
  fs::path res{};

  if (orig_file.size() >= 120) {  // Limit the length of local filenames
    auto [filename, ext]{split_file_ext(orig_file)};
    filename = filename.substr(0, 115);
    // If we have a file extension from the split then use it
    res = (media_dir / (filename + (ext != "" ? "." + ext : "")));
  }
  else
    res = (media_dir / orig_file);

  return res;
}

//------------------------------------------------------------------------------

// Separate a filename and it's extension out
std::tuple<std::string, std::string> Page::split_file_ext(
    const std::string& filename) const
{
  // Search backwards to find the last '.'
  auto last_dot_pos{filename.rfind('.')};

  if (last_dot_pos != string::npos) {
    const string file{filename.substr(0, last_dot_pos)};
    const string ext{filename.substr(last_dot_pos + 1, string::npos)};
    return make_tuple(file, ext);
  }
  else {  // dot not found
    return make_tuple(filename, "");
  }
}

//------------------------------------------------------------------------------

// First get the html response, then store both to disk
void Page::wrt_json_html()
{
  resp("html");

  write_file(page_nm, "json", targ_dir, raw_json);
  write_file(page_nm, "html", targ_dir, raw_html);
}

//------------------------------------------------------------------------------

// Overwrite various types of file data into a file
template<typename T>
void Page::write_file(const std::string& file_nm,
                      const std::string& ext,
                      const fs::path& targ_dir,
                      const T& file_dat) const
{
  const fs::path filename{file_nm + "." + ext};
  const fs::path filepath{targ_dir / filename};
  ofstream ofs(filepath.string());
  ofs << file_dat;
}

//------------------------------------------------------------------------------

Json::Value Page::json() const
{
  return jsondb.db;
}

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

// Parameterized Bump ctor
Bump::Bump(const int argc, char** const argv)
{
  std::tie(site_nm, board_nm, force_chk) = parse_args(argc, argv);
  arch_dir = ".";  // DESIGN: This can be added into the args list in the future
  init_arch();

  // Re-load the class' 'catalog' member field now that we have the needed data
  // properly available
  catalog = Page{site_nm, board_nm, "catalog", board_dir};

  chk_version();

  chk_load_bumps();
}

//------------------------------------------------------------------------------

// Parse the user-supplied arguments for the program
std::tuple<std::string, std::string, bool> Bump::parse_args(const int argc,
                                                            char** const argv)
{
  if (argc < 3) {
    cerr << "\nprogram argument error\n"
         << "usage: " << argv[0] << "  site_name  board_name\n";

    std::exit(-1);
  }

  const string site_nm{argv[1]};
  const string board_nm{argv[2]};

  // Optional 3rd flag argument to force full re-checking of the archive
  bool force_chk{false};
  if (argv[3]) {
    istringstream bool_arg{argv[3]};

    if (bool_arg.str() == "true")
      bool_arg >> std::boolalpha >> force_chk;
    else if (bool_arg.str() == "1")
      bool_arg >> force_chk;
    else {
      cerr << "\n'force check' flag argument error\n"
           << "usage: 'true' or '1'\n";

      std::exit(-1);
    }
  }

  return make_tuple(site_nm, board_nm, force_chk);
}

//------------------------------------------------------------------------------

// Initialize the archive in general
void Bump::init_arch()
{
  std::tie(board_dir, thrds_dir) = init_arch_dirs();

  cfg_fpath = (board_dir / ".archbot.config");
  if (!fs::exists(cfg_fpath))
    ofstream cfg_ofs{cfg_fpath.string()};

  cfg_json = rd_json_file(cfg_fpath);
}

//------------------------------------------------------------------------------

// Initialize the archive's basic directory structure within the filesystem
std::tuple<fs::path, fs::path> Bump::init_arch_dirs()
{
  const fs::path site_dir{site_nm};
  fs::create_directory(arch_dir / site_dir);
  const fs::path board_dir{site_dir / board_nm};
  fs::create_directory(board_dir);
  const fs::path thrds_dir{board_dir / "threads"};
  fs::create_directory(thrds_dir);
  const fs::path static_dir{board_dir / ".static"};
  fs::create_directory(static_dir);

  return make_tuple(board_dir, thrds_dir);
}

//------------------------------------------------------------------------------

// Load and parse a json file
Json::Value Bump::rd_json_file(const fs::path& filepath)
{
  ifstream json_file{filepath.string()};
  return parse_json(json_file);
}

//------------------------------------------------------------------------------

// Take a raw json source and parse it into a json db
template<typename T> Json::Value Bump::parse_json(T& json_src)
{
  Json::Reader reader{};
  Json::Value json_dat{};
  reader.parse(json_src, json_dat);
  return json_dat;
}

//------------------------------------------------------------------------------

// Check the version number of the config, resetting the file if outdated
void Bump::chk_version()
{
  // DESIGN: string 'less than' here now?
  //  -no need to nuke the file w/o a good reason in the future, as was the case
  //   w/ v0.1b vs. v0.1a (no 'subj_id_dir's in the "threads" array elements)
  bool vrsn_is_dated{cfg_json["version"] != bump_vrsn};

  if (vrsn_is_dated) {  // this will force a rebuild of the .archbot.config file
    ofstream cfg_ofs{cfg_fpath.string()};
    cfg_json.clear();
    cfg_json["version"] = bump_vrsn;
    force_chk = true;
  }
}

//------------------------------------------------------------------------------

// Check if we have any bumps, then compare live and archive threads together
void Bump::chk_load_bumps()
{
  // First check if we have any bumps at all
  load_live_thrds();
  cmp_last_bumps();

  // If so, proceed with pre-processing
  if (force_chk || have_bumps) {
    catalog.wrt_json_html();
    find_all_bumps();
  }
}

//------------------------------------------------------------------------------

// Load the live_thrds vector
void Bump::load_live_thrds()
{
  for (const auto& thrd : catalog.json())
    live_thrds.emplace_back(make_tuple(thrd["threadId"].asUInt(),
                                       thrd["subject"].asString(),
                                       thrd["lastBump"].asString()));
}

//------------------------------------------------------------------------------

// Compare the live board's 'last known bump' vs. the archive's
void Bump::cmp_last_bumps()
{
  // First thing, sort out the 'lastBump's from the board
  const auto live_bumps_srtd{sort_live_bumps()};
  const auto live_last_bump{live_bumps_srtd[0]};  // latest bump will be 0-th

  const auto cfg_last_bump{cfg_json["last_known_bump"].asString()};

  if (live_last_bump != cfg_last_bump) {  // We have a new last known bump
    have_bumps = true;
    cfg_json["last_known_bump"] = live_last_bump;
  }
  else
    have_bumps = false;
}

//------------------------------------------------------------------------------

// Sort the 'lastBump's from the live_thrds, descending. (newest first)
std::vector<std::string> Bump::sort_live_bumps()
{
  vector<string> live_bumps{};

  for (const auto& thrd : live_thrds)
    live_bumps.emplace_back(get<2>(thrd));  // bump field is 2-th

  sort(begin(live_bumps), end(live_bumps), std::greater<string>());

  return live_bumps;
}

//------------------------------------------------------------------------------

// Collect up all the newly-bumped thread IDs
void Bump::find_all_bumps()
{
  sort_live_thrds();
  load_cfg_thrds();

  cmp_all_bumps();
}

//------------------------------------------------------------------------------

// Sort the live_thrds vector by IDs
void Bump::sort_live_thrds()
{
  sort(begin(live_thrds),
       end(live_thrds),
       [](const auto& thrda, const auto& thrdb) {
         return get<0>(thrda) < get<0>(thrdb);  // 'id' is 0-th tuple element
       });
}

//------------------------------------------------------------------------------

// Load the pre-sorted cfg_thrds vector
void Bump::load_cfg_thrds()
{
  // The "threads" json array loaded from the .archbot.config is already sorted
  // by ID, so simply load the vector.
  for (const auto& thrd : cfg_json["threads"]) {
    cfg_thrds.emplace_back(make_tuple(thrd["threadId"].asUInt(),
                                      thrd["subject"].asString(),
                                      thrd["lastBump"].asString(),
                                      thrd["subj_id_dir"].asString()));
  }
}

//------------------------------------------------------------------------------

// Compare all the bumps from the live board and the archive, storing the
// difference between them (ie, keep the newly bumped IDs)
void Bump::cmp_all_bumps()
{
  // DESIGN: What about the case for thread deletions?
  //  - Dealing with deletions may require confirmation of IDs as well.

  // Ensure the cfg_thrds size is identical, in case live_thrds has new threads.
  const auto live_sz{live_thrds.size()};
  const auto cfg_sz{cfg_thrds.size()};
  const bool have_diff{live_sz != cfg_sz};
  if (have_diff)
    cfg_thrds.resize(live_sz);

  // Compare the live and config 'lastBump's side-by-side.
  for (size_t i{0}; i < live_sz; ++i) {
    const auto& [live_id, live_subj, live_bump]{live_thrds[i]};
    const auto& cfg_bump{get<2>(cfg_thrds[i])};  // 'bump' is 2-th tuple element

    if (force_chk || (live_bump != cfg_bump))
      bump_ids.emplace_back(live_id);
  }

  if (have_diff)               // Now remove any extraneous elements that were
    cfg_thrds.resize(cfg_sz);  // added just to prepare for the comparision loop
}

//------------------------------------------------------------------------------

std::string Bump::site() const
{
  return site_nm;
}

std::string Bump::name() const
{
  return board_nm;
}

bool Bump::has_bumps() const
{
  return have_bumps;
}

std::vector<uint> Bump::bumps() const
{
  return bump_ids;
}

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

// Return the thread vector's iterator if the ID is found inside the element
template<typename T> auto Bump::locate_id(const uint id, T& thrd_vec)
{
  return find_if(begin(thrd_vec), end(thrd_vec), [id](const auto& thrd) {
    // 'id' is in the 0-th position of the tuple
    return get<0>(thrd) == id;
  });
}

//------------------------------------------------------------------------------

// For any threads that are bumped, save all the files out (as needed), creating
// custom-named directories for any new threads.
void Bump::save_bump_thrds()
{
  if (force_chk)
    cerr << "\nChecking all threads...\n\n";

  // Iteratively check if each bumped thread already has a special directory
  // yet, and create one if not (that is, its a new thread and needs a folder).
  for (const auto bump_id : bump_ids) {
    string dir{};

    auto cfg_it = locate_id(bump_id, cfg_thrds);
    if (!force_chk &&
        (cfg_it != end(cfg_thrds))) {  // Thread's directory already created
      auto& [id, subj, cfg_bump, cfg_dir]{*cfg_it};
      dir = cfg_dir;

      // Get the live thread's bump and store it back into the config data
      const auto live_it = locate_id(bump_id, live_thrds);
      const auto& live_bump{get<2>(*live_it)};  // 'bump' is 2-th tuple element
      cfg_bump = live_bump;
    }
    //
    else {  // ID not found (ie, have a new thread) so need to create new dir

      // Get the new thread's data and create a custom dir from it's subj & id
      const auto live_it = locate_id(bump_id, live_thrds);
      const auto& [id, subj, bump]{*live_it};
      dir = add_thrd_dir(subj, bump_id);

      // Preserve the new thread's data, including the new dir path
      cfg_thrds.emplace_back(make_tuple(bump_id, subj, bump, dir));
    }

    // Store each bumped thread's Page for processing.
    bump_thrds.emplace_back(Page{site_nm, board_nm, to_string(bump_id), dir});
  }

  // Save all the bumped thread's files as needed.
  for (auto& thrd : bump_thrds) {
    thrd.grab_thrd_files();
    thrd.wrt_json_html();
  }

  update_cfg();
}

//------------------------------------------------------------------------------

// Store the updated thread's data back into the config's json.
void Bump::update_cfg()
{
  cfg_json["threads"].clear();
  for (const auto& [id, subj, bump, dir] : cfg_thrds) {
    Json::Value elem{};
    elem["threadId"] = id;
    elem["subject"] = subj;
    elem["lastBump"] = bump;
    elem["subj_id_dir"] = dir;

    cfg_json["threads"].append(elem);
  }

  write_cfg();
}

//------------------------------------------------------------------------------

// Write the config's json back out into the .archbot.config file
void Bump::write_cfg()
{
  // DESIGN: Just use an ofstream here now?
  fstream cfg_ofs{cfg_fpath.string(), ios::out};  // DESIGN: | ios::trunc ?
  fast_wrt_strm(cfg_ofs, cfg_json);
}

//------------------------------------------------------------------------------

// Write json out to a file using no whitespace (fastwrite)
void Bump::fast_wrt_strm(std::fstream& io_strm, const Json::Value& json)
{
  // Prepare the stream for writing
  io_strm.clear();
  io_strm.seekp(0);

  Json::FastWriter writer;
  io_strm << writer.write(json);
}

//------------------------------------------------------------------------------

// Given a thread's subject & id, create a custom directory for it under the
// general threads directory
fs::path Bump::add_thrd_dir(const std::string& subj, const uint id)
{
  const auto clean_subj{make_dir_name(subj)};
  const auto id_pad{make_id_pad(id)};
  const fs::path subj_pad{clean_subj + "_" + id_pad};
  const fs::path thrd_dir{thrds_dir / subj_pad};
  fs::create_directory(thrd_dir);

  return thrd_dir;
}

//------------------------------------------------------------------------------

// Clean up the thread's subject suitable for use as a portable directory name
std::string Bump::make_dir_name(const std::string& subj)
{
  string res{subj};

  const vector<pair<char, char>> bad_goods{{' ', '_'},  // intentionally first
                                           {'&', 'n'},
                                           {'?', ' '},
                                           {'!', ' '},
                                           {'^', ' '},
                                           {'$', ' '},
                                           {'*', ' '},
                                           {'=', ' '},
                                           {'/', ' '},
                                           {'\\', ' '},
                                           {'\0', ' '},
                                           {'\n', ' '},
                                           {'(', ' '},
                                           {')', ' '},
                                           {'.', ' '},
                                           {',', ' '},
                                           {':', ' '},
                                           {';', ' '}};
  for (const auto bad_good : bad_goods)
    zap_bad_char(res, bad_good);

  const vector<string> bad_strings{"nquot", "napos"};  // '&' already xformd 'n'
  for (const auto& bad_string : bad_strings)
    zap_bad_string(res, bad_string);

  zap_bad_string(res, " ");

  // Limit the length of local directory names (44 + '_12345' == 50 total)
  if (res.size() > 44)
    res = res.substr(0, 44);

  return res;
}

//------------------------------------------------------------------------------

// Replace bad chars with a good char
void Bump::zap_bad_char(std::string& in_out,
                        const std::pair<char, char> bad_good)
{
  const auto [bad, good]{bad_good};
  std::transform(begin(in_out), end(in_out), begin(in_out), [=](char ch) {
    return ch == bad ? good : ch;
  });
}

//------------------------------------------------------------------------------

// Remove bad sub-strings
void Bump::zap_bad_string(std::string& in_out, const std::string& bad)
{
  using str_sz_t = string::size_type;
  str_sz_t bad_len = bad.length();
  const auto str_end{string::npos};

  for (str_sz_t i = in_out.find(bad); i != str_end; i = in_out.find(bad)) {
    in_out.erase(i, bad_len);
  }
}

//------------------------------------------------------------------------------

// Pad the ID out to n digits, using zeros
std::string Bump::make_id_pad(const uint id, const uint n)
{
  std::stringstream res{};
  res << std::setfill('0') << std::setw(n) << id;
  return res.str();
}

// Copyright (2019)
// License (MIT) https://opensource.org/licenses/MIT
